package backend.opt;

import backend.component.RiscvBlock;
import backend.component.RiscvFunction;
import backend.component.RiscvInstr;
import backend.component.RiscvModule;
import backend.instruction.RiscvBinary;
import backend.instruction.RiscvBranch;
import backend.instruction.RiscvCall;
import backend.instruction.RiscvFld;
import backend.instruction.RiscvFlw;
import backend.instruction.RiscvFmv;
import backend.instruction.RiscvFsd;
import backend.instruction.RiscvFsw;
import backend.instruction.RiscvJ;
import backend.instruction.RiscvLd;
import backend.instruction.RiscvLw;
import backend.instruction.RiscvMv;
import backend.instruction.RiscvSd;
import backend.instruction.RiscvSw;
import backend.operand.RiscvImm;
import backend.operand.RiscvOperand;
import backend.operand.RiscvReg;
import ir.value.BasicBlock;
import tools.DoubleNode;

import java.util.ArrayList;

public class BackPeepHole {
    private final RiscvModule riscvModule;

    public BackPeepHole(RiscvModule riscvModule) {
        this.riscvModule = riscvModule;
    }

    public void process() {
        boolean finished = false;
        while (!finished)
        {
            finished = peephole();
            // TODO 数据流窥孔
            //finished &= dataFlowPeephole();
        }
    }

    private boolean peephole() {
        // 任何优化都不会发生，那么才叫做 finished
        boolean finished = true;
        for (RiscvFunction riscvFunction : riscvModule.getFunctions().values()) {
            // 正序遍历blocks
            for (RiscvBlock block : riscvFunction.getBlocks()) {
                for (DoubleNode<RiscvInstr> insNode = block.getRiscvInstrs().getHead();
                     insNode != null; insNode = insNode.getSucc()) {
                    finished &= moveSameRegOptimize(insNode);
                    finished &= addiOptimize(insNode);
                    finished &= storeLoadOptimize(insNode);
                }
            }

            // 倒序遍历blocks
            for (int i = riscvFunction.getBlocks().size() - 1; i >= 0; i--) {
                RiscvBlock block = riscvFunction.getBlocks().get(i);
                for (DoubleNode<RiscvInstr> insNode = block.getRiscvInstrs().getHead();
                     insNode != null; insNode = insNode.getSucc()) {
                    finished &= moveUselessOptimize(insNode); // 必须和 moveSameRegOptimize 分开
                }
                finished &= jumpOptimize(riscvFunction, block);
            }
        }
        return finished;
    }

    private boolean moveSameRegOptimize(DoubleNode<RiscvInstr> insNode) {
        boolean finished = true;
        if (insNode.getContent() instanceof RiscvMv) {
            // 将 mv rs, rs 删掉
            if (insNode.getContent().getDefReg() ==
                    (insNode.getContent().getOperands().get(0))) {
                insNode.getParentList().removeNode(insNode);
                finished = false;
            }
        }
        return finished;
    }

    private boolean moveUselessOptimize(DoubleNode<RiscvInstr> insNode) {
        boolean finished = true;
        if (insNode.getContent() instanceof RiscvMv) {
            // 将 mv rs, rs 删掉
            if (insNode.getContent().getDefReg() ==
                    (insNode.getContent().getOperands().get(0))) {
                insNode.getParentList().removeNode(insNode);
                finished = false;
            }

            // 连续两条指令 mv a1, a2 和 mv a1, a3 则删掉第一条指令
            DoubleNode<RiscvInstr> preNode = insNode.getPred();
            if (preNode != null && preNode.getContent() instanceof RiscvMv
                    && insNode.getContent().getDefReg().equals(preNode.getContent().getDefReg())) {
                insNode.getParentList().removeNode(preNode);
                finished = false;
            }
        }
        return finished;
    }

    private boolean addiOptimize(DoubleNode<RiscvInstr> insNode) {
        boolean finished = true;
        if (insNode.getContent() instanceof RiscvBinary) {
            RiscvBinary ins = (RiscvBinary) insNode.getContent();
            if (ins.getInstType().equals(RiscvBinary.RiscvBinaryType.addi)) {
                if (((RiscvImm) ins.getOperands().get(1)).getValue() == 0) {
                    // 将 addi rs, rs, 0 删掉
                    if (ins.getDefReg().equals(ins.getOperands().get(0))) {
                        insNode.getParentList().removeNode(insNode);
                        finished = false;
                    } else { // 将 addi rd, rs, 0 改成 mv rd, rs
                        DoubleNode<RiscvInstr> newNode = new DoubleNode<>
                                (new RiscvMv((RiscvReg) ins.getOperands().get(0),
                                        ins.getDefReg()));
                        insNode.getParentList().insertBeforeNode(insNode, newNode);
                        insNode.getParentList().removeNode(insNode);
                        finished = false;
                    }
                }
            } else if (ins.getInstType().equals(RiscvBinary.RiscvBinaryType.addiw)) {
                if (((RiscvImm) ins.getOperands().get(1)).getValue() == 0) {
                    // 将 addi rs, rs, 0 删掉
                    if (ins.getDefReg().equals(ins.getOperands().get(0))) {
                        insNode.getParentList().removeNode(insNode);
                        finished = false;
                    } else { // 将 addi rd, rs, 0 改成 mv rd, rs
                        DoubleNode<RiscvInstr> newNode = new DoubleNode<>
                                (new RiscvMv((RiscvReg) ins.getOperands().get(0),
                                        ins.getDefReg()));
                        insNode.getParentList().insertBeforeNode(insNode, newNode);
                        insNode.getParentList().removeNode(insNode);
                        finished = false;
                    }
                }
            }
        }
        return finished;
    }

    private boolean storeLoadOptimize(DoubleNode<RiscvInstr> insNode) {
        boolean finished = true;
        // 将 store a, memory; load b, sameMemory 改成 move b, a
        DoubleNode<RiscvInstr> preNode = insNode.getPred();
        if (preNode != null) {
            if ((insNode.getContent() instanceof RiscvLd
                    && preNode.getContent() instanceof RiscvSd
                    && insNode.getContent().getOperands().get(0)
                    .equals(preNode.getContent().getOperands().get(1))
                    && insNode.getContent().getOperands().get(1)
                    .equals(preNode.getContent().getOperands().get(2)))
                    || (insNode.getContent() instanceof RiscvLw
                    && preNode.getContent() instanceof RiscvSw
                    && insNode.getContent().getOperands().get(0)
                    .equals(preNode.getContent().getOperands().get(1))
                    && insNode.getContent().getOperands().get(1)
                    .equals(preNode.getContent().getOperands().get(2)))) {
                DoubleNode<RiscvInstr> newNode = new DoubleNode<>
                        (new RiscvMv((RiscvReg) preNode.getContent().getOperands().get(0),
                                insNode.getContent().getDefReg()));
                insNode.getParentList().insertBeforeNode(preNode, newNode);
                insNode.getParentList().removeNode(preNode);
                insNode.getParentList().removeNode(insNode);
                finished = false;
            } else if ((insNode.getContent() instanceof RiscvFld
                    && preNode.getContent() instanceof RiscvFsd
                    && insNode.getContent().getOperands().get(0)
                    .equals(preNode.getContent().getOperands().get(1))
                    && insNode.getContent().getOperands().get(1)
                    .equals(preNode.getContent().getOperands().get(2))) ||
                    (insNode.getContent() instanceof RiscvFlw
                            && preNode.getContent() instanceof RiscvFsw
                            && insNode.getContent().getOperands().get(0)
                            .equals(preNode.getContent().getOperands().get(1))
                            && insNode.getContent().getOperands().get(1)
                            .equals(preNode.getContent().getOperands().get(2)))) {
                DoubleNode<RiscvInstr> newNode = new DoubleNode<>
                        (new RiscvFmv((RiscvReg) preNode.getContent().getOperands().get(0),
                                insNode.getContent().getDefReg()));
                insNode.getParentList().insertBeforeNode(preNode, newNode);
                insNode.getParentList().removeNode(preNode);
                insNode.getParentList().removeNode(insNode);
                finished = false;
            }
        }
        return finished;
    }

    private boolean jumpOptimize(RiscvFunction riscvFunction, RiscvBlock block) {
        boolean finished = true;
        // 例如：在 blocklable1 的最后 j blocklable2， 那么合并两个block
        // 【前提：全局没有其他 j blocklable2 和 br blocklable2 了】
        DoubleNode<RiscvInstr> insNode = block.getRiscvInstrs().getTail();
        if (insNode.getContent() instanceof RiscvJ ||
                (insNode.getContent() instanceof RiscvBranch &&
                        riscvFunction.getBlocks().indexOf(block) + 1
                                == riscvFunction.getBlocks()
                                .indexOf(insNode.getContent().getOperands().get(2)))) {
            RiscvBlock defBlock = (RiscvBlock) insNode.getContent().getOperands().get(0);
            if (riscvFunction.getBlocks().contains(defBlock)
                    && riscvFunction.getBlocks().indexOf(block) + 1
                    == riscvFunction.getBlocks().indexOf(defBlock)) {// defBlock 是紧接着 block 的
                boolean signal = true;
                for (RiscvBlock riscvBlock : riscvFunction.getBlocks()) {
                    if (!riscvBlock.equals(block)) {
                        if (riscvBlock.getRiscvInstrs().getTail()
                                .getContent() instanceof RiscvJ
                                && riscvBlock.getRiscvInstrs().getTail().getContent()
                                .getOperands().get(0).equals(defBlock)) {
                            signal = false;
                            break;
                        }
                        if (riscvBlock.getRiscvInstrs().getTail()
                                .getContent() instanceof RiscvBranch
                                &&
                                riscvBlock.getRiscvInstrs().getTail().getContent().getOperands()
                                        .get(2).equals(defBlock)) {
                            signal = false;
                            break;
                        }
                    }
                }
                if (signal) {
                    var thisInsts = insNode.getParentList();
                    insNode.getParentList().removeNode(insNode);
                    thisInsts.addListBefore(defBlock.getRiscvInstrs());
                    riscvFunction.getBlocks().remove(defBlock);
                }
            }
        }
        return finished;
    }
}
